#  Copyright (c) 2020, Ayush Pandey. All rights reserved.
#  See LICENSE file in the project root directory for details.

from sympy import Symbol  # type: ignore
import numpy as np  # type: ignore
import pytest  # type: ignore

from autoreduce.system import System
from autoreduce.utils import get_reducible


@pytest.fixture
def system():
    """
    This method gets executed before every test. It sets up a test CRN:
    2A + B (k1)<->(k2) C, C (k3)--> D
    """
    A = Symbol("A")
    B = Symbol("B")
    C = Symbol("C")
    D = Symbol("D")
    k1 = Symbol("k1")
    k2 = Symbol("k2")
    k3 = Symbol("k3")

    params = [k1, k2, k3]
    # States:
    x = [A, B, C, D]

    # ODE in Sympy for the given test CRN with mass-action kinetics
    f = [
        -k1 * A**2 * B + k2 * C,
        -k1 * A**2 * B + k2 * C,
        k1 * A**2 * B - k2 * C - k3 * C,
        k3 * C,
    ]
    init_cond = np.ones(len(x))
    C = None
    g = None
    h = None
    u = None
    input_values = None
    params_values = [2, 4, 6]
    system = System(
        x,
        f,
        params=params,
        x_init=init_cond,
        params_values=params_values,
        C=C,
        g=g,
        h=h,
        u=u,
        input_values=input_values,
    )
    return system


@pytest.fixture
def reducible_system(system):
    """
    Pytest fixture that returns a reducible system
    """
    reducible_system = get_reducible(system)
    return reducible_system


def test_get_reduced_model(reducible_system):
    """
    This function creates a reducible System object
    that can be used to create reduced models given
    the x_hat (the list of states in reduced model).
    All other states are collapsed to be at quasi-steady
    state and both the reduced and the collapsed models
    are returned.
    """
    x_hat = []
    assert isinstance(reducible_system, System)
    reduced_system, collapsed_system = (
        reducible_system.solve_timescale_separation(x_hat)
    )
    if reduced_system is not None:
        assert isinstance(reduced_system, System)
    if collapsed_system is not None:
        assert isinstance(collapsed_system, System)



"""Local sensitivity analysis module."""

import numpy as np  # type: ignore
from scipy.integrate import solve_ivp, odeint  # type: ignore
from sympy import lambdify  # type: ignore

from autoreduce import utils
from .system import System


class SSM(System):
    """
    Class that computes local sensitivity
    analysis coefficients for the given Model using a numerical
    approximation method discussed in
    doi: https://doi.org/10.1016/0021-9991(76)90007-3
    Uses numerical method to find sensitivity analysis matrix (SSM).
    Both the Jacobian matrix and the Z matrix are estimated using 4th
    order central difference as given in the paper.
    """

    def __init__(
        self,
        x,
        f,
        params=None,
        C=None,
        g=None,
        h=None,
        u=None,
        params_values=None,
        x_init=None,
        timepoints=None,
    ):
        super().__init__(x, f, params, C, g, h, u, params_values, x_init)
        if timepoints is None:
            timepoints = []
        else:
            self.timepoints = timepoints

        return

    def compute_Zj(self, x, j, **kwargs):
        """
        Compute Z_j, i.e. df/dp_j at a particular
        timepoint k for the parameter p_j.
        Returns a vector of size n x 1.
        Use mode = 'accurate' for this object attribute
        to use accurate computations using numdifftools.
        """
        # if 'mode' in kwargs:
        #     if kwargs.get('mode') == 'accurate':
        #         del kwargs['mode']
        #         return self.sensitivity_to_parameter(x, j, **kwargs)
        # initialize Z
        Z = np.zeros(self.n)
        P_holder = self.params_values
        # For each state
        for i in range(self.n):
            P = P_holder
            F = np.zeros((4, 1))  # For 4th order difference
            h = P[j] * 0.01  # Small parameter for this parameter
            # Gets O(4) central difference on dfi/dpj
            if h != 0:
                P[j] = P_holder[j] + 2 * h
                f = self.evaluate(self.f, x, P)
                F[0] = f[i]
                P[j] = P_holder[j] + h
                f = self.evaluate(self.f, x, P)
                F[1] = f[i]
                P[j] = P_holder[j] - h
                f = self.evaluate(self.f, x, P)
                F[2] = f[i]
                P[j] = P_holder[j] - 2 * h
                f = self.evaluate(self.f, x, P)
                F[3] = f[i]
                # Store approx. dfi/dpj into Z
                Z[i] = (-F[0] + 8 * F[1] - 8 * F[2] + F[3]) / (12 * h)
                if Z[i] == np.inf:
                    Z[i] = 1
                elif Z[i] == np.nan:
                    Z[i] = 0

        return Z

    def compute_J(self, x, **kwargs):
        """
        Compute the Jacobian J = df/dx at a timepoint k.
        Returns a matrix of size n x n.
        Use mode = 'accurate' for this object attribute
        to use accurate computations using numdifftools.
        """
        if "fun" in kwargs:
            fun = kwargs.get("fun")
        else:
            fun = self.f
        if "var" in kwargs:
            var = kwargs.get("var")
        else:
            var = x
        # initialize J
        J = [[0] * self.n] * len(var)
        # Change this variable name P to params_values or something
        P = self.params_values
        if "set_params_as" in kwargs:
            set_params_as = kwargs["set_params_as"]
        else:
            set_params_as = None
        u = self.u
        if "mode" in kwargs:
            if kwargs.get("mode") == "accurate":
                del kwargs["mode"]
                try:
                    import numdifftools as nd
                except ImportError:
                    raise ValueError(
                        "The package numdifftools is not"
                        + "installed for this method to work."
                    )
                fun_l = lambdify((self.x, self.params), fun)

                def fun_ode(t, x, params):
                    y = fun_l(x, params)
                    return np.array(y)

                jfun = nd.Jacobian(lambda x: fun_ode(0, x, P), **kwargs)
                return jfun(x)
        # store the variable with respect to
        # which we approximate the differentiation (df/dvar)
        X = var
        for i in range(self.n):
            for j, _ in enumerate(var):
                # F = np.zeros( (4,1) )
                F = [None] * 4
                h = 0.01
                # Gets O(4) central difference on dfi/dvarj
                if h != 0:
                    var = X
                    var[j] = X[j] + 2 * h
                    f = self.evaluate(
                        fun, var, P, u, set_params_as=set_params_as
                    )
                    F[0] = f[i]
                    var[j] = X[j] + h
                    f = self.evaluate(
                        fun, var, P, u, set_params_as=set_params_as
                    )
                    F[1] = f[i]
                    var[j] = X[j] - h
                    f = self.evaluate(
                        fun, var, P, u, set_params_as=set_params_as
                    )
                    F[2] = f[i]
                    var[j] = X[j] - 2 * h
                    f = self.evaluate(
                        fun, var, P, u, set_params_as=set_params_as
                    )
                    F[3] = f[i]
                    # Store approvar. dfi/dvarj into J
                    J[i][j] = (-1 * F[0] + 8 * F[1] - 8 * F[2] + F[3]) / (
                        12 * h
                    )
                    # print(J[i,j])
                    # if J[i,j] == np.inf:
                    #     J[i,j] = 1
                    # elif J[i,j] == np.nan:
                    #     J[i,j] = 0
        if set_params_as is None:
            J = np.array(J, dtype=float)
        return J

    def compute_SSM(self, normalize=False, **kwargs):
        """
        Returns the sensitivity coefficients
        S_j for each parameter p_j.
        The sensitivity coefficients are written
        in a sensitivity matrix SSM of size
        len(timepoints) x len(params) x n
        If normalize argument is true,
        the coefficients are normalized by
        the nominal value of each paramneter.
        Use mode = 'accurate' for this object attribute
        to use accurate computations using numdifftools.
        """
        if "mode" in kwargs:
            if kwargs.get("mode") == "accurate_SSM":
                return self.solve_extended_ode(**kwargs)

        def sens_func(t, x, J, Z):
            # forms ODE to solve for sensitivity coefficient S
            dsdt = J @ x + Z
            return dsdt

        P = self.params_values
        S0 = np.zeros(self.n)  # Initial value for S_i
        SSM = np.zeros((len(self.timepoints), len(P), self.n))
        # solve for all x's in timeframe set by timepoints
        system_obj = self.get_system()
        sol = utils.get_ODE(system_obj, self.timepoints).solve_system().T
        xs = sol
        xs = np.reshape(xs, (len(self.timepoints), self.n))
        self.xs = xs
        # Solve for SSM at each time point
        for k, _ in enumerate(self.timepoints):
            # print('for timepoint',self.timepoints[k])
            timepoints = self.timepoints[0 : k + 1]  # noqa: E203
            if len(timepoints) == 1:
                continue
            # get the jacobian matrix
            J = self.compute_J(xs[k, :], **kwargs)
            # Solve for S = dx/dp for all x and all P
            # (or theta, the parameters) at time point k
            for j in range(len(P)):
                utils.printProgressBar(
                    int(j + k * len(P)),
                    len(self.timepoints) * len(P) - 1,
                    prefix="SSM Progress:",
                    suffix="Complete",
                    length=50,
                )
                # print('for parameter',P[j])
                # get the pmatrix
                Zj = self.compute_Zj(xs[k, :], j, **kwargs)

                # solve for S
                def sens_func_ode(t, x):
                    return sens_func(t, x, J, Zj)

                sol = odeint(sens_func_ode, S0, timepoints, tfirst=True)
                S = sol
                S = np.reshape(S, (len(timepoints), self.n))
                SSM[k, j, :] = S[k, :]
        self.SSM = SSM
        if normalize:
            SSM = (
                self.normalize_SSM()
            )  # Identifiablity was estimated using an normalized SSM
        return SSM

    def normalize_SSM(self):
        """
        Returns normalized sensitivity coefficients.
        Multiplies each sensitivity coefficient with
        the corresponding parameter p_j
        Divides the result by the corresponding state to
        obtain the normalized coefficient that is returned.
        """
        SSM_normalized = np.zeros(np.shape(self.SSM))
        for j, _ in enumerate(self.params_values):
            for i in range(self.n):
                SSM_normalized[:, j, i] = np.divide(
                    self.SSM[:, j, i] * self.params_values[j], self.xs[:, i]
                )
        self.SSM_normalized = SSM_normalized
        return SSM_normalized

    def get_system(self):
        return System(
            self.x,
            self.f,
            self.params,
            self.C,
            self.g,
            self.h,
            self.u,
            self.params_values,
            self.x_init,
        )

    """ Code contributed by Sam Clamons below """

    def sensitivity_to_parameter(self, x, j, **kwargs):
        """
        Calculates the response of each derivative (defined by ode) to changes
        in a single parameter (the jth one) at point x.

        keyword argument options?:
            ode_sol - An OdeSolution object holding a continuously-interpolated
                        solution for ode.
            ode_jac - Jacobian of the ode, as calculated by
                        numdifftools.Jacobian.
            ode - The ODE for the system, of the form ode(t, x, params)
            params - A list of parameters to feed to ode.
            p - The index of the parameter to calculate sensitivities to.
            t_min - Starting time.
            t_max - Ending time.
        returns: An OdeSolution object representing the
        (continously-interpolated)
                    sensitivity of each variable to the specified
                    parameter over time.
        """
        return None
        # Build a scipy-integratable derivative-of-sensitivity function.
        # import numdifftools as nd
        # import copy
        # def dS_dt(t, s):
        #     xs = ode_sol(t)
        #     # Wrapper to let numdifftools calculate df/dp.
        #     def ode_as_parameter_call(param):
        #         call_params = copy.deepcopy(self.params_values)
        #         call_params[j] = self.params_values
        #         return ode(t, xs, call_params)
        #     df_dp = lambda xs: nd.Jacobian(ode_as_parameter_call)\
        # (xs).transpose()[:,0]
        #     return df_dp(params[j]) + np.matmul(ode_jac(xs), s)
        # sol = odeint(dS_dt, np.zeros(n_vars), self.timepoints, **kwargs)
        # return sol

    def solve_extended_ode(
        self,
        ode=None,
        params=None,
        t_min=None,
        t_max=None,
        init=None,
        method="RK45",
    ):
        """
        Augments an ODE system (as a scipy-integratable function) into an ODE
        representing the original ODE plus sensitivities, then solves them all.

        The key equation here is, for a system dx/dt = f(x, p, t),

        dS_j/dt = f_j + J*S_j

        where S_j is the vector of sensitivities of xs
        to parameter j, f_j is the
        vector of derivatives df/dp_j, and J is the Jacobian of f w.r.t. xs.

        params:
            ode - An ode to solve, with a signature ode(t, xs, parameters).
            params - a vector of parameters around which to
            calculate sensitivity.
            t_min - Starting time.
            t_max - Ending time.
            init - Initial conditions for the ode.
            method - ODE solving method, passed directly to
                        scipy.integrate.odeint.

        Returns: (x_sols, sensitivities)
            x_sols - An OdeSolution object with the
            solution to the original ODE.
                        Shape of a solution is (n_variables, n_times)
            sensitivities - An array of OdeSolution objects of  size (n_params)
                            holding sensitivities of each variable to each
                            parameter over time, as a continuous interpolation.
                            Shape of a solution is (n_variables, n_times)
        """
        import numdifftools as nd  # type: ignore

        n_params = len(params)

        # Solve ODE.
        def ode_func(t, xs):
            return ode(t, xs, params)

        ode_jac = nd.Jacobian(lambda x: ode_func(0, x))
        sol = solve_ivp(
            ode_func,
            (t_min, t_max),
            init,
            method=method,
            dense_output=True,
            jac=lambda t, x: ode_jac(x),
        )

        if sol.status != 0:
            raise ValueError(
                "In solve_extended_ode, solve_ivp failed with "
                " error message: " + sol.message
            )

        sensitivities = [None] * n_params
        for p in range(n_params):
            print(f"Solving sensitivity for parameter {p + 1}, {n_params}")
            # sensitivities[p] = self.sensitivity_to_parameter(
            #     sol.sol, ode_jac, ode, params, n_variables, p, t_min, t_max
            # )

        return (sol.sol, sensitivities)



#  Copyright (c) 2020, Ayush Pandey. All rights reserved.
#  See LICENSE file in the project root directory for details.

import warnings
import pytest  # type: ignore


@pytest.fixture
def system_1_setup():
    system1 = None
    return system1


@pytest.fixture
def system_2_setup():
    system2 = None
    return system2


def test_system_equality(system1=None, system2=None):
    """
    Test all properties of two systems for equality
    """
    if system1 is None and system2 is None:
        return
    elif system1 is None or system2 is None:
        warnings.warn("One of the System objects is None.")
    else:
        test_states()
        test_f()
        test_g()
        test_h()
        test_params()
        test_initial_conditions()
        test_params_values()
        test_C()


def test_states(system1=None, system2=None):
    if system1 is not None and system2 is not None:
        assert system1.x == system2.x


def test_f(system1=None, system2=None):
    if system1 is not None and system2 is not None:
        assert system1.f == system2.f


def test_g(system1=None, system2=None):
    if system1 is not None and system2 is not None:
        assert system1.g == system2.g


def test_h(system1=None, system2=None):
    if system1 is not None and system2 is not None:
        assert system1.h == system2.h


def test_params(system1=None, system2=None):
    if system1 is not None and system2 is not None:
        assert system1.params == system2.params


def test_params_values(system1=None, system2=None):
    if system1 is not None and system2 is not None:
        assert system1.params_values == system2.params_values


def test_initial_conditions(system1=None, system2=None):
    if system1 is not None and system2 is not None:
        if system1.x_init is not None and system2.x_init is not None:
            assert system1.x_init == system2.x_init


def test_C(system1=None, system2=None):
    if system1 is not None and system2 is not None:
        if system1.C is not None and system2.C is not None:
            assert system1.C == system2.C



#  Copyright (c) 2025, Ayush Pandey. All rights reserved.
#  See LICENSE file in the project root directory for details.

import pytest  # type: ignore
from autoreduce import System


def test_reduced_models(system, reducible_system):
    A, B, C, D = system.x
    k1, k2, k3 = system.params
    possible_reductions = reducible_system.get_all_combinations()
    for attempt in possible_reductions:
        answer_AB = [
            A**2 * B * k1 * k2 / (k2 + k3) - A**2 * B * k1,
            A**2 * B * k1 * k2 / (k2 + k3) - A**2 * B * k1,
        ]
        # On simplification, the answer is:
        answer_AB_eq = [
            -(A**2) * B * k1 * k3 / (k2 + k3),
            -(A**2) * B * k1 * k3 / (k2 + k3),
        ]
        try:
            solve_timescale_test(
                system,
                reducible_system,
                attempt_states=[A, B],
                mode="fail",
                answer=answer_AB,
            )
        except Exception:
            solve_timescale_test(
                system,
                reducible_system,
                attempt_states=[A, B],
                mode="fail",
                answer=answer_AB_eq,
            )

        solve_timescale_test(
            system,
            reducible_system,
            attempt_states=[A, D],
            mode="fail",
            answer=[0, 0],
        )
        solve_timescale_test(
            system,
            reducible_system,
            attempt_states=[A, C],
            mode="fail",
            answer=[0, -C * k3],
        )
        solve_timescale_test(
            system,
            reducible_system,
            attempt_states=[B, C, D],
            mode="success",
            answer=[0, -k3 * C, k3 * C],
        )
        solve_timescale_test(
            system,
            reducible_system,
            attempt_states=[A, C, D],
            mode="success",
            answer=[0, -k3 * C, k3 * C],
        )
        # solve_timescale_test(attempt_states = [A, B, C],
        #                          mode = 'success',
        #                          answer = [-k1 * A**2 * B,
        #                                        -k1 * A**2 * B,
        #                                        k1 * A**2 * B])
        solve_timescale_test(
            system,
            reducible_system,
            attempt_states=[C, D],
            mode="success",
            answer=[-k3 * C, k3 * C],
        )
        answer_ABD = [
            k1 * k2 * A**2 * B / (k2 + k3) - k1 * A**2 * B,
            k1 * k2 * A**2 * B / (k2 + k3) - k1 * A**2 * B,
            k1 * k3 * A**2 * B / (k2 + k3),
        ]
        # On simplification, the answer is:
        answer_ABD_eq = [
            -(A**2) * B * k1 * k3 / (k2 + k3),
            -(A**2) * B * k1 * k3 / (k2 + k3),
            A**2 * B * k1 * k3 / (k2 + k3),
        ]
        try:
            solve_timescale_test(
                system,
                reducible_system,
                attempt_states=[A, B, D],
                mode="success",
                answer=answer_ABD,
            )
        except Exception:
            solve_timescale_test(
                system,
                reducible_system,
                attempt_states=[A, B, D],
                mode="success",
                answer=answer_ABD_eq,
            )


def solve_timescale_test(
    system, reducible_system, attempt_states=None, mode=None, answer=None
):
    if mode == "fail":
        with pytest.warns(Warning, match="Solve time-scale separation failed"):
            reduced_system, collapsed_system = get_reduced_model_helper(
                system, reducible_system, x_hat=attempt_states
            )
            if reduced_system is not None:
                print(reduced_system.f)
                assert reduced_system.f == answer
                assert collapsed_system.x == [
                    x for x in system.x if x not in attempt_states
                ]
    elif mode == "success":
        reduced_system, collapsed_system = get_reduced_model_helper(
            system, reducible_system, x_hat=attempt_states
        )
        assert answer is not None
        assert reduced_system is not None
        assert reduced_system.f == answer
        # Test slow (retained) states
        assert reduced_system.x == attempt_states
        # Test fast (collapsed) states
        assert collapsed_system.x == [
            x for x in system.x if x not in attempt_states
        ]


def get_reduced_model_helper(system, reducible_system, x_hat=None):
    _ = system
    if x_hat is None:
        x_hat = []
    assert isinstance(reducible_system, System)
    reduced_system, collapsed_system = (
        reducible_system.solve_timescale_separation(x_hat)
    )
    if reduced_system is not None:
        assert isinstance(reduced_system, System)
    if collapsed_system is not None:
        assert isinstance(collapsed_system, System)
    return reduced_system, collapsed_system



from autoreduce import ode
from autoreduce import local_sensitivity
from autoreduce import model_reduction


def get_ODE(system_obj, timepoints, **kwargs):
    """
    For the given timepoints,
    create an ODE class object for this System object.
    """
    ode_obj = ode.ODE(
        system_obj.x,
        system_obj.f,
        C=system_obj.C,
        g=system_obj.g,
        h=system_obj.h,
        u=system_obj.u,
        params=system_obj.params,
        params_values=system_obj.params_values,
        x_init=system_obj.x_init,
        timepoints=timepoints,
        **kwargs,
    )
    return ode_obj


def solve_ODE_SSM(system_obj, timepoints_ode, timepoints_ssm, **kwargs):
    """
    For the given timepoints,
    returns the full solution
    (states, sensitivity coefficients, outputs)
    """
    ode = system_obj.get_ODE(timepoints_ode)
    ssm = system_obj.get_SSM(timepoints_ssm)
    x_sol = ode.solve_system().y
    y = system_obj.C @ x_sol
    Ss = ssm.compute_SSM()
    return x_sol, y, Ss


def get_SSM(system_obj, timepoints, **kwargs):
    """
    For the given timepoints,
    create an SSM class object for this System object.
    """
    ssm_obj = local_sensitivity.SSM(
        system_obj.x,
        system_obj.f,
        g=system_obj.g,
        C=system_obj.C,
        h=system_obj.h,
        u=system_obj.u,
        params=system_obj.params,
        params_values=system_obj.params_values,
        x_init=system_obj.x_init,
        timepoints=timepoints,
        **kwargs,
    )
    return ssm_obj


def get_reducible(
    system_obj, timepoints_ode=None, timepoints_ssm=None, **kwargs
):
    red_obj = model_reduction.Reduce(
        system_obj.x,
        system_obj.f,
        C=system_obj.C,
        params=system_obj.params,
        g=system_obj.g,
        h=system_obj.h,
        u=system_obj.u,
        params_values=system_obj.params_values,
        x_init=system_obj.x_init,
        parameter_dependent_ic=system_obj.parameter_dependent_ic,
        ic_parameters=system_obj.ic_parameters,
        timepoints_ode=timepoints_ode,
        timepoints_ssm=timepoints_ssm,
        **kwargs,
    )
    return red_obj


def reduce_utils(reduce_obj, **kwargs):
    reduce_utils_obj = model_reduction.ReduceUtils(
        reduce_obj.x,
        reduce_obj.f,
        C=reduce_obj.C,
        params=reduce_obj.params,
        g=reduce_obj.g,
        h=reduce_obj.h,
        u=reduce_obj.u,
        params_values=reduce_obj.params_values,
        x_init=reduce_obj.x_init,
        timepoints_ode=reduce_obj.timepoints_ode,
        timepoints_ssm=reduce_obj.timepoints_ssm,
        error_tol=reduce_obj.error_tol,
        nstates_tol=reduce_obj.nstates_tol,
    )
    return reduce_utils_obj


def get_ode_solutions(system_obj, timepoints, **kwargs):
    x_sol = get_ODE(system_obj, timepoints, **kwargs).solve_system().T
    return x_sol


def printProgressBar(
    iteration, total, prefix="", suffix="", decimals=1, length=100, fill="█"
):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals
                                  in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(
        100 * (iteration / float(total))
    )
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    print("\r%s |%s| %s%% %s" % (prefix, bar, percent, suffix), end="\r")
    # Print New Line on Complete
    if iteration == total:
        print()



#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

import numpy as np  # type: ignore
from scipy.integrate import odeint  # type: ignore

from autoreduce.ode import ODE
from autoreduce.system import System
from autoreduce.utils import get_ODE


def test_ode_objects(system):
    import numpy as np  # type: ignore

    timepoints = np.linspace(0, 10, 100)
    ode_object = ODE(
        system.x,
        system.f,
        system.params,
        system.C,
        system.g,
        system.h,
        system.u,
        params_values=[2, 4, 6],
        x_init=np.ones(4),
        input_values=system.input_values,
        timepoints=timepoints,
    )
    ode_object_same = get_ODE(system, timepoints=timepoints)
    assert isinstance(ode_object, ODE)
    assert isinstance(ode_object, System)
    assert ode_object.f == ode_object_same.f
    assert isinstance(ode_object.get_system(), System)

    # Compare values instead of objects
    ode_system = ode_object.get_system()
    assert ode_system.x == system.x
    assert ode_system.f == system.f
    assert ode_system.params == system.params
    assert ode_system.params_values == system.params_values
    assert np.array_equal(ode_system.x_init, system.x_init)
    assert ode_system.C == system.C
    assert ode_system.g == system.g
    assert ode_system.h == system.h
    assert ode_system.u == system.u
    assert ode_system.input_values == system.input_values


def test_ode_solutions(system):
    """
    Solve the ODE manually and solve using
    ODE class to compare the two solutions
    """
    timepoints = np.linspace(0, 10, 100)
    params_values = [2, 4, 6]
    ode_object = ODE(
        system.x,
        system.f,
        system.params,
        system.C,
        system.g,
        system.h,
        system.u,
        params_values=params_values,
        x_init=np.ones(4),
        input_values=system.input_values,
        timepoints=timepoints,
    )
    solutions = ode_object.solve_system()

    assert isinstance(solutions, np.ndarray)

    # Solve manually:
    def scipy_odeint_func(x, t):
        k1, k2, k3 = params_values
        A, B, C, D = x
        return np.array(
            [
                -k1 * A**2 * B + k2 * C,
                -k1 * A**2 * B + k2 * C,
                k1 * A**2 * B - k2 * C - k3 * C,
                k3 * C,
            ]
        )

    solutions_manual = odeint(scipy_odeint_func, y0=np.ones(4), t=timepoints)
    assert (solutions == solutions_manual).all()



import numpy as np  # type: ignore

from autoreduce import load_ODE_model
from autoreduce import System
from autoreduce.ode import ODE
from autoreduce.local_sensitivity import SSM


# ## System object attributes


# Create symbolic objects
x, f, P = load_ODE_model(2, 2)
f[0] = -x[0] ** 2 + P[0] * x[1]
f[1] = -P[1] * x[1]
print("The states : {0}".format(x))
print("The dynamics f : {0}".format(f))
print("The dynamics P : {0}".format(P))
C = np.array([[0, 1]]).tolist()


# System class
sys = System(x, f, params=P, C=C)


# ## Solve the System using the ODE subclass


# Solve ODE from System

timepoints = np.linspace(0, 20, 100)
sys.params_values = [2, 4]
sys.x_init = [0, 10]
sys_ode = ODE(
    sys.x,
    sys.f,
    params=sys.params,
    params_values=sys.params_values,
    C=sys.C,
    x_init=sys.x_init,
    timepoints=timepoints,
)
solution = sys_ode.solve_system()


# ## Local sensitivity analysis tools for System using the SSM subclass


# Solve for sensitivity analysis from System

timepoints = np.linspace(0, 20, 10)
sys.params_values = [2, 4]
sys.x_init = [0, 10]
sys_ssm = SSM(
    sys.x,
    sys.f,
    params=sys.params,
    params_values=sys.params_values,
    C=sys.C,
    x_init=sys.x_init,
    timepoints=timepoints,
)
solution = sys_ssm.compute_SSM()

J = sys_ssm.compute_J([2, 1])
Z = sys_ssm.compute_Zj([2, 1], 1)
print("J = ", J)
print("Z = ", Z)
print("SSM = ", solution)



"""ODE class"""

import time  # type: ignore

import numpy as np  # type: ignore
from sympy import lambdify  # type: ignore
from scipy.integrate import odeint  # type: ignore

from .system import System


class ODE(System):
    """
    To solve the Model using scipy.odeint
    """

    def __init__(
        self,
        x,
        f,
        params=None,
        C=None,
        g=None,
        h=None,
        u=None,
        params_values=None,
        x_init=None,
        input_values=None,
        timepoints=None,
    ):
        super().__init__(
            x, f, params, C, g, h, u, params_values, x_init, input_values
        )
        if timepoints is None:
            timepoints = []
            self.timepoints = timepoints
        else:
            self.timepoints = timepoints
        return

    def solve_system(self, **kwargs):
        """
        Solve the System ODE for the given timepoints
        and initial conditions set for System.
        Other options passed to scipy.integrate.odeint.
        """
        if self.u is not None:
            print(self.u)
            return self.solve_system_with_inputs()

        fun = lambdify((self.x, self.params), self.f)

        def fun_ode(t, x, params):
            y = fun(x, params)
            return np.array(y)

        if "timing" in kwargs:
            timing = kwargs.get("timing")
            del kwargs["timing"]
        else:
            timing = False
        start_time = time.time()
        sol = odeint(
            lambda t, x: fun_ode(t, x, self.params_values),
            self.x_init,
            self.timepoints,
            tfirst=True,
            **kwargs,
        )
        end_time = time.time()
        if timing:
            print("Time taken for this ODE call : ", end_time - start_time)
        self.sol = sol
        return sol

    def solve_system_with_inputs(self, **kwargs):
        """
        Solve the system in self with inputs (u) values known.
        """
        if self.u is None:
            raise ValueError(
                "Use solve_system() if there are no inputs in the System."
            )
        f_g = [fi + gi * self.u for fi, gi in zip(self.f, self.g)]
        fun = lambdify((self.x, self.u, self.params), f_g)

        def fun_ode(t, x, u, params):
            y = fun(x, u, params)
            return np.array(y)

        if "timing" in kwargs:
            timing = kwargs.get("timing")
            del kwargs["timing"]
        else:
            timing = False

        start_time = time.time()
        sol = odeint(
            lambda t, x: fun_ode(t, x, self.input_values, self.params_values),
            self.x_init,
            self.timepoints,
            tfirst=True,
            **kwargs,
        )

        end_time = time.time()
        if timing:
            print("Time taken for this ODE call : ", end_time - start_time)
        self.sol = sol
        return sol

    def get_system(self):
        """Returns the system object of the ODE.

        Returns:
            System: The system object
        """
        return System(
            self.x,
            self.f,
            self.params,
            self.C,
            self.g,
            self.h,
            self.u,
            self.params_values,
            self.x_init,
        )



import libsbml
import numpy as np


def create_sbml_model(
    compartment_id="default",
    time_units="second",
    extent_units="mole",
    substance_units="mole",
    length_units="metre",
    area_units="square_metre",
    volume_units="litre",
    volume=1e-6,
    model_id=None,
    **kwargs,
):
    """Creates an SBML Level 3 Version 2 model
    with some fixed standard settings.
    Refer to python-libsbml for more information on SBML API.
    :param compartment_id:
    :param time_units:
    :param extent_units:
    :param substance_units:
    :param length_units:
    :param area_units:
    :param volume_units:
    :param volume:
    :param model_id:
    :return:  the SBMLDocument and the Model object as a tuple
    """
    document = libsbml.SBMLDocument(3, 2)
    model = document.createModel()
    if model_id is None:
        model_id = "autoreduce_" + str(np.random.randint(1, 1e6))
    model.setId(model_id)
    model.setName(model_id)
    # Define units for area
    unitdef = model.createUnitDefinition()
    unitdef.setId("square_metre")
    unit = unitdef.createUnit()
    unit.setKind(libsbml.UNIT_KIND_METRE)
    unit.setExponent(2)
    unit.setScale(0)
    unit.setMultiplier(1)

    # Set up required units and containers
    model.setTimeUnits(time_units)  # set model-wide time units
    model.setExtentUnits(extent_units)  # set model units of extent
    model.setSubstanceUnits(substance_units)  # set model substance units
    model.setLengthUnits(length_units)  # area units
    model.setAreaUnits(area_units)  # area units
    model.setVolumeUnits(volume_units)  # default volume unit
    compartment = model.createCompartment()
    compartment.setId(compartment_id)
    compartment.setVolume(volume)
    compartment.setConstant(True)
    return document, model


def add_species(model, species, initial_condition=0):
    """Add species to SBML model

    Args:
        model (libsbml.Model): SBML model to add the species to
        species (str): Species string to add to model
        initial_condition (float, optional):Initial
        condition of species. Defaults to 0.
    Returns:
        libsbml.Model: SBML Model object
    """
    default_compartment = model.getCompartment(0)
    sbml_species = model.createSpecies()
    sbml_species.setName(species)
    sbml_species.setId(species)
    sbml_species.setCompartment(default_compartment.getId())
    sbml_species.setConstant(False)
    sbml_species.setBoundaryCondition(False)
    sbml_species.setHasOnlySubstanceUnits(False)
    sbml_species.setSubstanceUnits("mole")
    sbml_species.setInitialConcentration(initial_condition)
    return model


def add_reaction(
    model: libsbml.Model,
    species: str,
    kinetic_law: str,
    reaction_id: str,
    all_species: list,
):
    """Add reaction to given model

    Args:
        model (libsbml.Model): SBML model object
        species (str): Species name
        kinetic_law (str): Kinetic law for the species in the ODE
                           with sympy Symbols
        reaction_id (str): Reaction identifier string
        all_species (list): List of all species in System

    Returns:
        libsbml.Model : SBML Model object
    """
    sbml_reaction = model.createReaction()
    sbml_reaction.setId(reaction_id)
    sbml_reaction.setName(reaction_id)
    sbml_reaction.setReversible(False)
    # Create product
    product = sbml_reaction.createProduct()
    product.setSpecies(species)
    product.setConstant(False)
    # Create modifiers
    modifier_species = []
    for element in kinetic_law.free_symbols:
        if str(element) in all_species and str(element) != species:
            modifier_species.append(str(element))
    for modifier_id in modifier_species:
        modifier = sbml_reaction.createModifier()
        modifier.setSpecies(modifier_id)
    # Create kinetic law
    ratelaw = sbml_reaction.createKineticLaw()
    str_kinetic_law = str(kinetic_law)
    str_kinetic_law = str_kinetic_law.replace("**", "^")
    math_ast = libsbml.parseL3Formula(str_kinetic_law)
    flag = ratelaw.setMath(math_ast)
    if not flag == libsbml.LIBSBML_OPERATION_SUCCESS or math_ast is None:
        raise ValueError(
            f"Could not write the rate law for reaction to SBML."
            f"Check the ODE functions of species {species}."
        )
    return model


def add_parameters(
    model: libsbml.Model, all_parameters: list, all_values: list
):
    """Adds global parameters to the SBML Model

    Args:
        model (libsbml.Model): SBML Model to add global parameters to
        all_parameters (List): List of all parameter name symbols
        all_values (List): List of corresponding parameter values

    Returns:
        libsbml.Model: Updated SBML model object
    """
    for name, value in zip(
        all_parameters,
        all_values,
    ):
        if model.getParameter(name) is None:
            param = model.createParameter()
            param.setId(name)
            param.setConstant(True)
            param.setValue(value)
        else:
            param = model.getParameter(name)
    return model



"""Model reduction"""

import warnings
from itertools import combinations

from sympy import Symbol, solve, Eq  # type: ignore
import sympy  # type: ignore
import numpy as np  # type: ignore

from scipy.linalg import solve_lyapunov, block_diag  # type: ignore
from scipy.linalg import eigvals, norm  # type: ignore

from .system import System
from autoreduce import utils


class Reduce(System):
    """
    The class can be used to compute the various
    possible reduced models for the System object
    and then find out the best reduced
    model choice using doi : https://doi.org/10.1101/640276
    """

    def __init__(
        self,
        x,
        f,
        params=None,
        C=None,
        g=None,
        h=None,
        u=None,
        params_values=None,
        x_init=None,
        timepoints_ode=None,
        timepoints_ssm=None,
        error_tol=None,
        nstates_tol=None,
        nstates_tol_min=None,
        **kwargs,
    ):
        super().__init__(
            x, f, params, C, g, h, u, params_values, x_init, **kwargs
        )
        self.f_hat = []  # Should be a list of Sympy objects
        if nstates_tol is None:
            self.nstates_tol = self.n - 1
        else:
            self.nstates_tol = nstates_tol
        if nstates_tol_min is None:
            self.nstates_tol_min = 1
        else:
            self.nstates_tol_min = nstates_tol_min
        if error_tol is None:
            self.error_tol = 1e6
        else:
            self.error_tol = error_tol
        if timepoints_ode is None:
            self.timepoints_ode = np.linspace(0, 100, 100)
        else:
            self.timepoints_ode = timepoints_ode
        if timepoints_ssm is None:
            self.timepoints_ssm = np.linspace(0, 100, 10)
        else:
            self.timepoints_ssm = timepoints_ssm
        self.results_dict = {}
        self.x_c = []
        self.x_sol = None
        self.x_sol2 = None
        self.full_ssm = None
        return

    def get_output_states(self):
        if self.C is None and self.h is None:
            return []
        outputs = np.dot(np.array(self.C), np.array(self.x))  # Get y = C*x
        if type(outputs) is not np.ndarray:
            outputs = [outputs]
        else:
            outputs = outputs.tolist()
        output_symbols = [list(i.free_symbols) for i in outputs]
        output_states = [
            item for sublist in output_symbols for item in sublist
        ]
        return output_states

    def get_all_combinations(self):
        """
        Combinatorially create sets of all states
        that can be reduced in self.all_reductions.
        In addition, returns the possible reductions
        list after removing the sets that
        contain states involved in the outputs.
        """
        from itertools import combinations

        possible_reductions = []
        n = self.n
        for i in range(n):
            if i != n - 1:
                comb = combinations(list(range(n)), i + 1)
                possible_reductions.append(list(comb))
        possible_reductions = [
            list(item) for sublist in possible_reductions for item in sublist
        ]
        self.all_combinations = [i for i in possible_reductions]
        output_states = self.get_output_states()
        restart = False
        x = self.x
        for attempt in self.all_combinations:
            states_attempt = [x[i] for i in attempt]
            if (
                not len(set(states_attempt).intersection(set(output_states)))
                == len(output_states)
                or len(attempt) > self.nstates_tol
            ):
                restart = True
            if restart:
                possible_reductions.remove(attempt)
                restart = False

        # Remove state(s) that consist of input(s)
        if self.u is not None:
            for i, _ in enumerate(self.g):
                if self.g[i] != 0:
                    if i in possible_reductions:
                        # This index state variable should
                        # not be in possible_reductions list
                        possible_reductions.remove(i)
        return possible_reductions

    def get_T(self, attempt):
        non_attempt = [i for i in range(self.n) if i not in attempt]
        T = np.zeros((self.n, self.n))
        n_hat = len(attempt)
        n = self.n
        n_c = n - n_hat
        T1 = np.zeros((self.n, n_hat))
        T2 = np.zeros((self.n, n_c))
        # For x_hat
        for ni in range(0, n_hat):
            set_T = False
            for i in range(n):
                if i in attempt and not set_T:
                    T[ni, i] = 1
                    attempt.remove(i)
                    set_T = True
        # For x_c
        for ni in range(n_hat, n):
            set_T = False
            for i in range(n):
                if i in non_attempt and not set_T:
                    T[ni, i] = 1
                    non_attempt.remove(i)
                    set_T = True
        T1 = T[0:n, 0:n_hat]
        T2 = T[0:n, n_hat : n + 1]  # noqa: E203
        return T, T1, T2

    def get_error_metric_with_input(self, reduced_sys):
        """
        Returns the error defined as the 2-norm of y - y_hat.
        y = Cx and y_hat = C_hat x_hat OR
        y = h(x, P), y_hat = h_hat(x_hat, P).
        Important : What is the input?
        """
        reduced_ode = utils.get_ODE(reduced_sys, self.timepoints_ode)
        x_sol, _, _ = self.get_solutions()
        y = self.C @ x_sol
        x_sols_hat = reduced_ode.solve_system().T
        reduced_sys.x_sol = x_sols_hat
        y_hat = np.array(reduced_sys.C) @ np.array(x_sols_hat)
        if np.shape(y) == np.shape(y_hat):
            e = np.linalg.norm(y - y_hat)
        else:
            raise ValueError(
                "The output dimensions must be the same for"
                + "reduced and full model. Choose C and C_hat accordingly"
            )
        if np.isnan(e):
            print("The error is NaN, something wrong...continuing.")
        return e

    def get_error_metric(self, reduced_sys):
        # Give option for get_error_metric(sys1, sys2)
        """
        Returns the error defined as the 2-norm of y - y_hat.
        y = Cx and y_hat = C_hat x_hat OR
        y = h(x, P), y_hat = h_hat(x_hat, P)
        """
        reduced_ode = utils.get_ODE(reduced_sys, self.timepoints_ode)
        x_sol, _, _ = self.get_solutions()
        y = self.C @ x_sol
        x_sols_hat = reduced_ode.solve_system().T
        reduced_sys.x_sol = x_sols_hat
        y_hat = np.array(reduced_sys.C) @ np.array(x_sols_hat)
        if np.shape(y) == np.shape(y_hat):
            e = np.linalg.norm(y - y_hat)
        else:
            raise ValueError(
                "The output dimensions must be the same for"
                + "reduced and full model. Choose C and C_hat accordingly"
            )
        if np.isnan(e):
            print("The error is NaN, something wrong...continuing.")
        return e

    def get_robustness_metric_with_input(self, reduced_sys):
        return

    def get_robustness_metric(self, reduced_sys, **kwargs):
        # Create an option so the default way this is
        # done is given two systems compute robustness metric.
        # Implementing Theorem 2
        if "method" in kwargs:
            method = kwargs.get("method")
        else:
            method = "direct"
        timepoints_ssm = self.timepoints_ssm
        _, x_sols, full_ssm = self.get_solutions()
        S = full_ssm.compute_SSM()
        self.S = S
        reduced_ssm = utils.get_SSM(reduced_sys, timepoints_ssm)
        x_sols_hat = (
            utils.get_ODE(reduced_sys, timepoints_ssm).solve_system().T
        )
        x_sols = np.reshape(x_sols, (len(timepoints_ssm), self.n))
        x_sols_hat = np.reshape(
            x_sols_hat, (len(timepoints_ssm), reduced_sys.n)
        )
        Se = np.zeros(len(self.params_values))
        S_hat = reduced_ssm.compute_SSM()
        reduced_sys.S = S_hat
        S_bar = np.concatenate((S, S_hat), axis=2)
        S_bar = np.reshape(
            S_bar,
            (
                len(timepoints_ssm),
                self.n + reduced_sys.n,
                len(self.params_values),
            ),
        )
        C_bar = np.concatenate((self.C, -1 * reduced_sys.C), axis=1)
        C_bar = np.reshape(
            C_bar, (np.shape(self.C)[0], (self.n + reduced_sys.n))
        )
        weighted_Se_sum = 0
        P_prev = None
        prev_time = None
        if method == "bound":
            for j, _ in enumerate(self.params_values):
                S_metric_max = 0
                sens_max = 0
                max_eig_P = 0
                max_eig_dot_P = 0
                for k in range(len(self.timepoints_ssm)):
                    curr_time = self.timepoints_ssm[k]
                    J = full_ssm.compute_J(x_sols[k, :])
                    J_hat = reduced_ssm.compute_J(x_sols_hat[k, :])
                    J_bar = block_diag(J, J_hat)
                    # print(J)
                    # print(J_bar)
                    # if np.isnan(J).any() or np.isnan(J_hat).any()
                    # or np.isfinite(J).all() or np.isfinite(J_hat).all():
                    #     warnings.warn('NaN or
                    #                   inf found in Jacobians, continuing')
                    #     continue
                    P = solve_lyapunov(J_bar, -1 * C_bar.T @ C_bar)
                    eig_P = max(eigvals(P))
                    # if k == 0: # used when proof I thought s
                    #              aid that lambda_max_P
                    #              was at time 0 for IC term.
                    if max_eig_P < eig_P:
                        max_eig_P = eig_P
                    if k != 0:
                        dot_P = (P - P_prev) / (curr_time - prev_time)
                        eig_dot_P = max(eigvals(dot_P))
                        if max_eig_dot_P < eig_dot_P:
                            max_eig_dot_P = eig_dot_P
                    Z = full_ssm.compute_Zj(x_sols[k, :], j)
                    Z_hat = reduced_ssm.compute_Zj(x_sols_hat[k, :], j)
                    Z_bar = np.concatenate((Z, Z_hat), axis=0)
                    Z_bar = np.reshape(Z_bar, ((self.n + reduced_sys.n), 1))
                    S_metric = norm(Z_bar.T @ P @ S_bar[k, :, j])
                    if S_metric > S_metric_max:
                        S_metric_max = S_metric
                    sens_norm = norm(S_bar[k, :, j]) ** 2
                    if sens_norm > sens_max:
                        sens_max = sens_norm
                    P_prev = P
                    prev_time = curr_time
                    utils.printProgressBar(
                        int(j + k * len(self.params_values)),
                        len(timepoints_ssm) * len(self.params_values) - 1,
                        prefix="Robustness Metric Progress:",
                        suffix="Complete",
                        length=50,
                    )
                dot_P_term = (
                    max_eig_dot_P * len(reduced_ssm.timepoints) * sens_max
                )
                Se[j] = (
                    max_eig_P
                    + 2 * len(reduced_ssm.timepoints) * S_metric_max
                    + dot_P_term
                )
                weighted_Se_sum += self.params_values[j] * Se[j]
        elif method == "direct":
            for j in range(len(self.params_values)):
                Se[j] = norm(C_bar @ S_bar[:, :, j].T)
                weighted_Se_sum += self.params_values[j] * Se[j]
        err_norm = norm(self.get_error_metric(reduced_sys))
        R = 1 / (1 + (weighted_Se_sum / err_norm))
        reduced_sys.R = R
        reduced_sys.Se = Se
        return Se, R

    def get_invariant_manifold(self, reduced_sys):
        timepoints_ode = self.timepoints_ode
        x_c = reduced_sys.x_c
        fast_states = reduced_sys.fast_states
        x_hat = reduced_sys.x
        x_sols_hat = reduced_sys.get_ODE().solve_system().T
        x_sol_c = np.zeros((len(timepoints_ode), np.shape(x_c)[0]))
        # Get the collapsed states by substituting the solutions
        # into the algebraic relationships obtained
        for i in range(np.shape(x_sols_hat)[0]):
            # for each reduced variable (because collapsed variables are only
            # functions of reduced variables, algebraically)
            for k in range(len(x_sols_hat[:, i])):
                for j in range(len(x_c)):
                    subs_result = fast_states[j].subs(
                        x_hat[i], x_sols_hat[:, i][k]
                    )
                    if subs_result == fast_states[j]:
                        continue
                    elif isinstance(subs_result, sympy.Expr):
                        # continue substituting other variables,
                        # until you get a float
                        fast_states[j] = subs_result
                    else:
                        x_sol_c[k][j] = fast_states[j].subs(
                            x_hat[i], x_sols_hat[:, i][k]
                        )
        return x_sol_c

    def solve_timescale_separation(
        self, slow_states, fast_states=None, **kwargs
    ):
        """
        This function solves the time-scale separation
        problem for the System object passed through.
        Arguments:
        * slow_states: List of states assumed to
        have slow dynamics => retained in the reduced model.
        This is list of Sympy Symbol objects
        corresponding to the System.x list.
        * fast_states: List of states assumed to
        have fast dynamics => states that will be collapsed. Usually,
        this is automatically populated as all those
        states that are in System.x but not in slow_states.
        """
        # Get 'debug' keyword (if called in debug = True mode):
        if "debug" in kwargs:
            debug = kwargs.get("debug")
        else:
            debug = False
        # If slow_states is empty, then reduced model = given model
        # and collapsed model is None
        if not slow_states:
            return self.get_system(), None
        x, f, x_init = self.x, self.f, self.x_init

        len_slow_states = len(slow_states)
        x_hat = [None] * len_slow_states
        x_hat_init = [None] * len_slow_states
        f_hat = [None] * len_slow_states

        max_len_fast_states = len(x) - len(slow_states)
        x_c = [None] * max_len_fast_states
        x_c_init = [None] * max_len_fast_states
        f_c = [None] * max_len_fast_states
        # print('f_c',f_c)
        # Populate the list of states that will retained (x_hat)
        # and those that will be collapsed (x_c)
        x_hat = slow_states
        # Make sure fast_states (direct sum) slow_states is all of it.
        # Check if fast states are already provided as well:
        if fast_states:
            x_c = fast_states
        else:
            # If not, automatically fill it up.
            count_x_c = 0
            for i in x:
                if i not in slow_states:
                    x_c[count_x_c] = i
                    count_x_c += 1
            fast_states = x_c
        # Consistency check for slow and fast states:
        if len(slow_states) + len(fast_states) != len(self.x):
            raise RuntimeError(
                "Number of slow states plus number of fast"
                "states must equal the number of total states."
            )
        for state in slow_states:
            if state in fast_states:
                raise RuntimeError(
                    "Found a state that is both fast and slow!"
                    "Unfortunately, that is not yet possible in this reality."
                )
        # Now populate the default corresponding
        # f_c and f_hat dynamics from self.f
        # and inital conditions from self.x_init
        for i in x:
            state_index = x.index(i)
            if i in x_c:
                x_c_index = x_c.index(i)
                f_c[x_c_index] = f[state_index]
                if self.parameter_dependent_ic:
                    param_as_ic = self.ic_parameters[state_index]
                    value_ic = self.set_ic_from_params(  # noqa: F841
                        x_c_init, param_as_ic, x_c_index
                    )
                else:
                    x_c_init[x_c_index] = x_init[state_index]
            if i in x_hat:
                x_hat_index = x_hat.index(i)
                f_hat[x_hat_index] = f[state_index]
                if self.parameter_dependent_ic:
                    param_as_ic = self.ic_parameters[state_index]
                    value_ic = self.set_ic_from_params(  # noqa: F841
                        x_hat_init, param_as_ic, x_hat_index
                    )
                else:
                    x_hat_init[x_hat_index] = x_init[state_index]
        self.f_hat = f_hat
        self.f_c = f_c
        if debug:
            print("Reduced set of variables is", x_hat)
            print("f_hat = ", self.f_hat)
            print("Collapsed set of variables is", x_c)

        # Get the reduced (slow system) dynamics:
        # Check after substituting each solution into
        # f_hat whether resulting f_hat sympy ODE
        # has any remaining variables that should have been collapsed.
        loop_sanity = True
        count = 0
        solution_dict = {}
        while (
            sympy_variables_exist(
                ode_function=self.f_hat, variables_to_check=x_c
            )[0]
            and loop_sanity
        ):
            # print(sympy_solve_and_substitute(ode_function = self.f_hat,
            #                                  collapsed_states = x_c,
            #                                  collapsed_dynamics = self.f_c,
            #                                  solution_dict = solution_dict,
            #                                  debug = debug))
            self.f_hat, solution_dict, self.f_c = sympy_solve_and_substitute(
                ode_function=self.f_hat,
                collapsed_states=x_c,
                collapsed_dynamics=self.f_c,
                solution_dict=solution_dict,
                debug=debug,
            )
            if count > 2:
                warnings.warn(
                    "Solve time-scale separation failed."
                    "Check model consistency."
                )
                print(
                    f"Did not work to retain: {slow_states}"
                    " because either a collapsed state-variables appears"
                )
                print(" in the reduced model or a solution is not possible.")
                loop_sanity = False
                return None, None
            count += 1

        # Get the collapsed (fast system) dynamics to create collapsed_system
        for i, _ in enumerate(x_hat):
            for j, _ in enumerate(self.f_c):
                # The slow variables stay at steady state in the fast subsystem
                self.f_c[j] = self.f_c[j].subs(x_hat[i], x_hat_init[i])

        # Create C_hat
        C_hat = self.create_C_hat(x_hat)
        for index, _ in enumerate(f_hat):
            f_hat[index] = sympy.simplify(f_hat[index])
        for index, _ in enumerate(f_c):
            f_c[index] = sympy.simplify(f_c[index])
        reduced_sys = create_system(
            x_hat,
            self.f_hat,
            params=self.params,
            C=C_hat,
            params_values=self.params_values,
            x_init=x_hat_init,
        )
        fast_subsystem = create_system(
            x_c,
            self.f_c,
            params=self.params,
            params_values=self.params_values,
            x_init=x_c_init,
        )
        reduced_sys.fast_states = fast_states
        # If you got to here,
        print(f"Successful solution obtained with states: {reduced_sys.x}!")
        return reduced_sys, fast_subsystem

    def solve_timescale_separation_with_input(self, attempt_states):
        attempt = []
        for state in attempt_states:
            attempt.append(self.x.index(state))
        print("attempting to retain:", attempt)
        x_c = []
        fast_states = []
        f_c = []
        f_hat = []
        x_hat_init = []
        x_c_init = []
        x_hat = []
        x, f, g, u, x_init = self.x, self.f, self.g, self.u, self.x_init
        f_g = [fi + gi for fi, gi in zip(f, g)]
        for i in range(self.n):
            if i not in attempt:
                x_c.append(x[i])
                f_c.append(f_g[i])
                x_c_init.append(x_init[i])
            else:
                f_hat.append(f_g[i])
                x_hat.append(x[i])
                x_hat_init.append(x_init[i])
        # print('Reduced set of variables is', x_hat)
        # print('f_hat = ',f_hat)
        # print('Collapsed set of variables is', x_c)

        solved_states = []
        lookup_collapsed = {}
        for i, _ in enumerate(x_c):
            x_c_sub = solve(Eq(f_c[i], 0), x_c[i])
            lookup_collapsed[x_c[i]] = x_c_sub
            if len(x_c_sub) == 0:
                # print('Could not find solution for this collapsed variable :
                # {0} from {1}'.format(x_c[i], f_c[i]))
                fast_states.append([])
                continue
            elif len(x_c_sub) > 1:
                # print('Multiple solutions obtained.
                # Chooosing non-zero solution,
                # check consistency. The solutions are ', x_c_sub)
                for sub in x_c_sub:
                    if sub == 0:
                        x_c_sub.remove(0)
            else:
                for sym in x_c_sub[0].free_symbols:
                    if sym in solved_states and sym in x:
                        # print('The state {0} has been solved for but appears
                        # in the solution for the next variable,
                        # making the sub with
                        # {1} into the corresponding f_c
                        # and solving again should
                        # fix this.'.format(sym, lookup_collapsed[sym][0]))
                        f_c[i] = f_c[i].subs(sym, lookup_collapsed[sym][0])
                        # print('Updating old x_c_sub then')
                        x_c_sub = solve(Eq(f_c[i], 0), x_c[i])
                        if len(x_c_sub) > 1:
                            print("Multiple solutions obtained.")
                            print(
                                "Chooosing non-zero solution,"
                                "check consistency."
                            )
                            print(" The solutions are ", x_c_sub)
                            for sub in x_c_sub:
                                if sub == 0:
                                    x_c_sub.remove(0)
                        # print('with ',x_c_sub)
                        lookup_collapsed[x_c[i]] = x_c_sub
                    else:
                        solved_states.append(x_c[i])
                # print('Solved for
                # {0} to get {1}'.format(x_c[i], x_c_sub[0]))
                # This x_c_sub should not contain previously eliminated
                # variables otherwise circles continue
                fast_states.append(x_c_sub[0])

        for i in range(len(fast_states)):
            if fast_states[i] == []:
                continue
            for j in range(len(f_hat)):
                # print('Substituting {0} for
                # variable {1}
                # into f_hat{2}'.format(fast_states[i], x_c[i], j))
                f_hat[j] = f_hat[j].subs(x_c[i], fast_states[i])
                # print('f_hat = ',f_hat[j])
            for j in range(len(f_c)):
                # print('Substituting {0} for
                # variable {1} into f_c{2}'.format(fast_states[i], x_c[i], j))
                f_c[j] = f_c[j].subs(x_c[i], fast_states[i])
                # print('f_c = ',f_c[j])

        # Continue
        for i in range(len(x_hat)):
            for j in range(len(f_c)):
                # The slow variables stay at steady state in the fast subsystem
                f_c[j] = f_c[j].subs(x_hat[i], x_hat_init[i])

        # Create C_hat
        output_states = self.get_output_states()
        C_hat = np.zeros((np.shape(self.C)[0], np.shape(x_hat)[0]))
        is_output = 0
        for i in range(len(x_hat)):
            if x_hat[i] in output_states:
                is_output = 1
            for row_ind in range(np.shape(C_hat)[0]):
                C_hat[row_ind][i] = 1 * is_output

        # Create list of all free symbols
        flag = False
        free_symbols_all = []
        for fi in f_hat:
            fi = sympy.sympify(fi)
            for sym in fi.free_symbols:
                if sym not in free_symbols_all:
                    free_symbols_all.append(sym)
        bugged_states = []
        for syms in free_symbols_all:
            if syms not in x_hat + u + self.params:
                bugged_states.append(syms)
                flag = True
        if flag:
            warnings.warn("Check model consistency")
            print(
                f"The time-scale separation that retains states {attempt},\
                  does not work"
            )
            print(
                f"because the state-variables {bugged_states} \
                  appear in the reduced model"
            )
            # return None, None

        reduced_sys = create_system(
            x_hat,
            f_hat,
            params=self.params,
            C=C_hat,
            params_values=self.params_values,
            x_init=x_hat_init,
        )
        fast_subsystem = create_system(
            x_c,
            f_c,
            params=self.params,
            params_values=self.params_values,
            x_init=x_c_init,
        )
        reduced_sys.x_c = x_c
        reduced_sys.bugged_states = bugged_states
        reduced_sys.fast_states = fast_states
        return reduced_sys, fast_subsystem

    def create_C_hat(self, x_hat):
        """
        Returns C_hat matrix for the reduced system
        given the x_hat (reduced system state vector)
        and using the C matrix for the full system (if any)
        """
        if self.C is None:
            C_hat = None
        else:
            output_states = self.get_output_states()
            C_hat = np.zeros((np.shape(self.C)[0], np.shape(x_hat)[0]))
            is_output = 0
            for i in range(len(x_hat)):
                if x_hat[i] in output_states:
                    is_output = 1
                for row_ind in range(np.shape(C_hat)[0]):
                    C_hat[row_ind][i] = 1 * is_output
        return C_hat

    def get_conservation_laws(self, num_conservation_laws: int, **kwargs):
        """Finds sets of conserved species.
        Only linear combinations with coefficient = 1 supported.

        Args:
            num_conservation_laws (int): The null space of the
                                       stoichiometry matrix. In other words,
                                       the number of expected
                                       conservation laws.
        Returns:
            List of conserved species (list)
        """
        all_conserved_sets = []
        ode_list = [i for i in self.f if i != 0]
        for d in range(num_conservation_laws):
            curr_depth = d + 1
            for i in combinations(ode_list, curr_depth):
                sum_terms = 0
                for element in i:
                    if element == 0:
                        continue
                    sum_terms += element
                if sum_terms == 0:
                    conserved_species = []
                    for element in i:
                        for ode_i in range(len(self.f)):
                            if str(self.f[ode_i]) == str(element):
                                conserved_species.append(self.x[ode_i])
                    if len(conserved_species) <= 1:
                        continue
                    all_conserved_sets.append(conserved_species)
        if not all_conserved_sets:
            raise ValueError(
                "No conserved sets found. Try increasing the "
                "depth of search by increasing the number of "
                "possible conservation laws: num_conservation_laws"
            )
        return all_conserved_sets

    def setup_conservation_laws(
        self, total_quantities: dict, conserved_sets: list
    ):
        """Setup conservation laws and return a
        list of conservation laws where
        each conservation law is each sublist of
        conserved_sets equated to the corresponding
        total quantity in the total quantities dictionary.

        Args:
            total_quantities (dict): Dictionary with total quantities
                                     string keys and total value
            conserved_sets (list): A list of list consisting
            where each sublist is a
                                   set of species that are conserved
        Returns:
            conservation_laws (list): A list of conservation laws
        """
        # Setup conservation laws
        params = self.params
        params_values = self.params_values
        conservation_laws = []
        for conserved_set, tot in zip(conserved_sets, total_quantities.keys()):
            total_symbol = Symbol(tot)
            params.append(total_symbol)
            params_values.append(total_quantities[tot])
            law = 0
            for x in conserved_set:
                law += x
            law += -total_symbol
            conservation_laws.append(law)
        self.params = params
        self.params_values = params_values
        return conservation_laws

    def solve_conservation_laws(
        self,
        conservation_laws: list = None,
        total_quantities: dict = None,
        conserved_sets: list = None,
        states_to_eliminate: list = None,
        num_conservation_laws: int = 0,
        **kwargs,
    ):
        """User interface wrapper to find and set
        conservation laws for a given Reduce System object

        Args:
            conservation_laws (list, optional): A list consisting of
                                                conservation laws
                                                in the form: LHS - RHS.
                                                The RHS is assumed to be zero.
                                                If None is provided,
                                                then attempts to find
                                                conservation laws,
                                                if num_conservation_laws set.
            total_quantities (dict, optional): A dictionary of total
                                               quantities with keys
                                               consisting of strings of
                                               total quantities
                                               (RHS of conservation law)
                                               and a float value.
                                               If None provided, then a
                                               dict with parameter name
                                               is created from the state
                                               name appended by keyword
                                               "_total" and with zero value.
            conserved_sets (list of list, optional): A list of list where each
                                                     sublist consists
                                                     of species in System.x,
                                                     for which,
                                                     if corresponding elements
                                                     in System.f
                                                     are added would be equal
                                                     to zero.
                                                     If None is provided,
                                                     then attempts to find the
                                                     conserved_sets,
                                                     if num_conservation_laws
                                                     is set.
            states_to_eliminate (list, optional): A list of states to eliminate
                                                  from the set
                                                  of conserved species.
                                                  Each element in this
                                                  list must correspond to each
                                                  sublist in conserved_sets
                                                  and/or conservation_laws,
                                                  depending on
                                                  which is passed in.
                                                  If None is provided,
                                                  then creates a default
                                                  list of
                                                  states to eliminate from
                                                  variables in each law in
                                                  conservation_laws list.
            num_conservation_laws (int, optional): The dimension of the
                                                   nullspace of the
                                                   stoichiometry
                                                   matrix. In other words,
                                                   the number of expected
                                                   conservation laws.
                                                   Defaults to 0 but then
                                                   expects that either
                                                   conservation_laws or
                                                   conserved_sets is given.
        Returns:
            conserved_system (Reduce): The reduced system with
                                       conservation laws applied.

        """
        debug = kwargs.get("debug", False)
        if (
            num_conservation_laws == 0
            and conserved_sets is None
            and conservation_laws is None
        ):
            raise ValueError(
                "Must pass in something to set conservation laws! "
                "Either the list of conservation_laws, or"
                "number of conservation "
                "laws through num_conservation_laws "
                "argument or the conserved_sets list"
            )
        if (
            conservation_laws is None
            and num_conservation_laws == 0
            and conserved_sets is not None
        ):
            if conserved_sets:
                self.num_conservation_laws = len(conserved_sets)
                self.conserved_sets = conserved_sets
            else:
                raise ValueError("List of conserved sets must not be empty.")
        elif (
            conservation_laws is None
            and num_conservation_laws != 0
            and conserved_sets is None
        ):
            self.num_conservation_laws = num_conservation_laws
            self.conserved_sets = self.get_conservation_laws(
                self.num_conservation_laws
            )
        else:
            self.conserved_sets = conserved_sets

        if conservation_laws is None:
            if total_quantities is None:
                total_quantities = {}
                for c_set in self.conserved_sets:
                    total_quantities[str(c_set[0]) + "_total"] = 0
                self.total_quantities = total_quantities
            else:
                self.total_quantities = total_quantities
            self.conservation_laws = self.setup_conservation_laws(
                self.total_quantities, self.conserved_sets
            )
            print("Found conservation laws:", self.conservation_laws)
        else:
            self.conservation_laws = conservation_laws
        # Remove duplicate laws
        for law_i, law in enumerate(self.conservation_laws):
            list_conservation_laws = list(self.conservation_laws)
            list_conservation_laws.remove(law)
            if law in list_conservation_laws:
                if debug:
                    print(
                        "Found duplicate law {0} on index {1}."
                        "This will be removed. Check conservation_laws"
                        "attribute to confirm.".format(law, law_i)
                    )
                self.conservation_laws.remove(law)
        if self.conservation_laws is not None and states_to_eliminate is None:
            # Conservation laws are passed in as a list
            # but states_to_eliminate list is not available
            # Then, create it by choosing one variable from each law
            states_to_eliminate = []
            for law in self.conservation_laws:
                list_of_symbols_in_law = list(law.free_symbols)
                chosen_var = None
                index = 0
                while chosen_var is None:
                    if list_of_symbols_in_law[index] in self.x:
                        chosen_var = list_of_symbols_in_law[index]
                    index += 1
                    if index == len(list_of_symbols_in_law):
                        raise ValueError(
                            "No variable found in conservation"
                            "law {0} that can be eliminated".format(law)
                        )
                states_to_eliminate.append(chosen_var)
            self.states_to_eliminate = states_to_eliminate
            print("Choosing states to eliminate:", self.states_to_eliminate)
        else:
            self.states_to_eliminate = states_to_eliminate

        self.f = self.set_conservation_laws(
            conservation_laws=self.conservation_laws,
            states_to_eliminate=self.states_to_eliminate,
        )
        return self

    def set_conservation_laws(self, conservation_laws, states_to_eliminate):
        """
        From the conserved_quantities list,
        this method computes the expressions
        for each of the state indices in states_to_eliminate,
        and substitutes into the full model dynamics.
        Both lists should contain symbolic variables
        referencing states in self.f.
        Returns the dynamics self.f.

        Args:
            conservation_laws (list): List of conservation laws
            states_to_eliminate (list): List of Symbols of states
                                            to eliminate when applying the
                                            conservation laws

        Returns:
            Conserved ODE[list]: Conserved ODE as a list of expressions.
        """
        states_to_eliminate_new = []
        for state in states_to_eliminate:
            states_to_eliminate_new.append(self.x.index(state))
        states_to_eliminate = states_to_eliminate_new
        for i in range(len(states_to_eliminate)):
            state = self.x[states_to_eliminate[i]]
            state_sub = solve(Eq(conservation_laws[i], 0), state)
            for j in range(len(self.f)):
                self.f[j] = self.f[j].subs(state, state_sub[0])

        arr_x = np.array(self.x)
        self.x = np.delete(arr_x, states_to_eliminate).tolist()
        arr_f = np.array(self.f)
        self.f = np.delete(arr_f, states_to_eliminate).tolist()
        self.x_init = np.delete(self.x_init, states_to_eliminate).tolist()
        if self.parameter_dependent_ic:
            self.ic_parameters = np.delete(
                self.ic_parameters, states_to_eliminate
            ).tolist()
        self.C = np.delete(np.array(self.C), states_to_eliminate, axis=1)
        self.n = self.n - len(states_to_eliminate)
        return self.f

    def solve_approximations(self):
        pass

    def get_solutions(self):
        if self.x_sol is None:
            x_sol = utils.get_ode_solutions(
                self.get_system(), self.timepoints_ode
            )
            self.x_sol = x_sol
        if self.x_sol2 is None:
            x_sol2 = utils.get_ode_solutions(
                self.get_system(), self.timepoints_ssm
            )
            self.x_sol2 = x_sol2
        if self.full_ssm is None:
            full_ssm = utils.get_SSM(self.get_system(), self.timepoints_ssm)
            self.full_ssm = full_ssm
        return self.x_sol, self.x_sol2, self.full_ssm

    def reduce_simple(self, **kwargs):
        if "skip_numerical_computations" in kwargs:
            skip_numerical_computations = kwargs.get(
                "skip_numerical_computations"
            )
        else:
            skip_numerical_computations = False
        if "skip_error_computation" in kwargs:
            skip_error_computation = kwargs.get("skip_error_computation")
        else:
            skip_error_computation = False
        if "skip_robustness_computation" in kwargs:
            skip_robustness_computation = kwargs.get(
                "skip_robustness_computation"
            )
        else:
            skip_robustness_computation = False
        if self.u is not None:
            raise ValueError("For models with inputs use reduce_with_input.")
        results_dict = {}
        possible_reductions = self.get_all_combinations()
        if not len(possible_reductions):
            print("No possible reduced models found.")
            print(" Try increasing tolerance for number of states.")
            return
        for attempt in possible_reductions:
            if len(attempt) < self.nstates_tol_min:
                continue
            elif len(attempt) > self.nstates_tol:
                continue
            attempt_states = [self.x[i] for i in attempt]
            # Create reduced systems
            reduced_sys, fast_subsystem = self.solve_timescale_separation(
                attempt_states, **kwargs
            )
            if reduced_sys is None or fast_subsystem is None:
                continue
            if skip_numerical_computations:
                results_dict[reduced_sys] = None
            else:
                # Get metrics for this reduced system
                if skip_error_computation:
                    e = np.nan
                else:
                    e = self.get_error_metric(reduced_sys)
                if skip_robustness_computation:
                    Se = np.nan
                    R = np.nan
                else:
                    Se, R = self.get_robustness_metric(reduced_sys, **kwargs)
                results_dict[reduced_sys] = [e, Se, R]
        self.results_dict = results_dict
        return self.results_dict

    def reduce_with_input(self):
        if self.u is None:
            raise ValueError("For models with no inputs use reduce_simple")
        results_dict = {}
        possible_reductions = self.get_all_combinations()
        if not len(possible_reductions):
            print("No possible reduced models found.")
            print(" Try increasing tolerance for number of states.")
            return
        for attempt in possible_reductions:
            attempt_states = [self.x[i] for i in attempt]
            # Create reduced systems
            r_sys, f_sys = self.solve_timescale_separation_with_input(
                attempt_states
            )
            if r_sys is None or f_sys is None:
                continue
            # Get metrics for this reduced system
            e = self.get_error_metric_with_input(r_sys)
            Se, R = self.get_robustness_metric_with_input(r_sys)
            results_dict[r_sys] = [e, Se, R]
        self.results_dict = results_dict
        return self.results_dict

    def reduce_general(self):
        results_dict = {}
        possible_reductions = self.get_all_combinations()
        if not len(possible_reductions):
            print("No possible reduced models found.")
            print(" Try increasing tolerance for number of states.")
            return

        self.results_dict = results_dict
        return self.results_dict

    def compute_reduced_model(self):
        if self.C is not None and self.g is None:
            # Call y = Cx based model reduction
            print("Using model reduction algorithm with y = Cx")
            print(" linear output relationship and no inputs (g = 0).")
            self.results_dict = self.reduce_simple()
            return self.results_dict
        else:
            print("Using general model reduction algorithm")
            print(" with inputs and nonlinear output relationship")
            self.results_dict = self.reduce_general()
            return self.results_dict

    def get_system(self):
        return System(
            self.x,
            self.f,
            params=self.params,
            C=self.C,
            g=self.g,
            h=self.h,
            u=self.u,
            params_values=self.params_values,
            x_init=self.x_init,
        )


def sympy_variables_exist(ode_function, variables_to_check, **kwargs):
    """
    To check whether any variable in variables_to_check
    appears in any Sympy equation
    in the ode_function list of functions.
    Arguments:
    * ode_function: A list of Sympy functions
                    that need to be checked.
    * variables_to_check: A list of variables that
                          need to be checked for in the ode_function
    * kwargs:
    Returns a tuple:
    True, variables_that_appear:
        If any variable found, True is returned
        along with a list of variables that appear
    False, []:
        If no variable found, False is returned
        with a None (since no variables are found in ode_function).
    """
    flag = False
    all_free_symbols = []
    if "debug" in kwargs:
        debug = kwargs.get("debug")
    else:
        debug = False
    if debug:
        print(
            "In sympy_variables_exist. Checking for presence of ",
            variables_to_check,
        )
    for fi in ode_function:
        fi = sympy.sympify(fi)
        for sym in fi.free_symbols:
            if sym not in all_free_symbols:
                all_free_symbols.append(sym)

    variables_that_appear = []
    for sym in all_free_symbols:
        if sym in variables_to_check:
            variables_that_appear.append(sym)
            flag = True
    if flag and debug:
        print("Found! The following: ", variables_that_appear)
    return flag, variables_that_appear


def sympy_solve_and_substitute(
    ode_function,
    collapsed_states,
    collapsed_dynamics,
    solution_dict,
    debug=False,
):
    """A function to solve using sympy and substitute into the equation"""
    for s in collapsed_states:
        index = collapsed_states.index(s)
        f = collapsed_dynamics[index]
        if debug:
            print("In sympy_solve_and_substitute, solving for ", s)
            print("From ", f)
        solution_dict = sympy_get_steady_state_solutions(
            collapsed_variables=[s],
            collapsed_dynamics=[f],
            solution_dict=solution_dict,
            debug=debug,
        )
        if debug:
            print("Solution found: ", solution_dict)
            print("current state", s)
        if solution_dict[s] is None or len(solution_dict[s]) == 0:
            continue
        for func in ode_function:
            func = sympy.sympify(func)
            func_index = ode_function.index(func)
            updated_func = func.subs(s, solution_dict[s][0])
            ode_function[func_index] = updated_func
        for func in collapsed_dynamics:
            if func == f:
                continue
            func_index = collapsed_dynamics.index(func)
            updated_func = func.subs(s, solution_dict[s][0])
            collapsed_dynamics[func_index] = updated_func
        if debug:
            print("Updated f_hat now is ", ode_function)
        # print('returning', ode_function)
        # print('returning', solution_dict)
        # print('returning', collapsed_dynamics)
        # returned_tuple = (ode_function, solution_dict, collapsed_dynamics)
        # if len(returned_tuple) == 3:
        # print('WTF is happening')
    # return returned_tuple
    return (ode_function, solution_dict, collapsed_dynamics)


def sympy_get_steady_state_solutions(
    collapsed_variables, collapsed_dynamics, solution_dict=None, debug=False
):
    """
    Solve for each collapsed_variable from
    corresponding collapsed_dynamics one by one.
    Return the solutions as a lookup dictionary
    as a map for variables and their solutions.
    """
    if solution_dict is None:
        solution_dict = {}
    x_c = collapsed_variables
    f_c = collapsed_dynamics
    for i, _ in enumerate(x_c):
        x_c_sub = solve(Eq(f_c[i], 0), x_c[i])
        if x_c_sub is None or len(x_c_sub) == 0:
            print(f"Could not find solution for: {x_c[i]} from {f_c[i]}")
            warnings.warn(
                "Solve time-scale separation failed. Check model consistency."
            )
        elif len(x_c_sub) > 1:
            if debug:
                print(f"Multiple solutions obtained for {x_c[i]}.")
                print("Chooosing one non-zero solution, check consistency. ")
                print(f"The solutions are {x_c_sub}.")
                print(" Highly recommend manuallly solving for this")
                print(" variable first then try this function.")
            for sub in x_c_sub:
                if sub == 0:
                    x_c_sub.remove(0)
        elif not any(x_c_sub):
            warnings.warn(
                "Solve time-scale separation failed. Check model consistency."
            )
            if debug:
                warnings.warn(f"Zero solution(s) for: {x_c[i]} from {f_c[i]}.")
        # Search for variables existing in x_c_sub
        # that might have been solved for before:
        # flag, solved_vars = sympy_variables_exist(ode_function = x_c_sub,
        #                                           variables_to_check =
        #                                           list(solution_dict.keys()),
        #                                           debug = debug)
        # if debug and flag:
        #     print('Found variables while solving
        # that have already been solved:', solved_variables)
        # if flag:
        #     for solved_variable in solved_Vars:
        #         for sub_func in x_c_sub:
        #             i = x_c_sub.index(sub_func)
        #             sub_func = sub_func.subs(solved_variable,
        # solution_dict[solved_variable][0])
        #             x_c_sub[i] = sub_func
        # Store solution in a lookup dictionary:
        solution_dict[x_c[i]] = x_c_sub
    return solution_dict


class ReduceUtils(Reduce):
    """
    For various utility methods developed
    on top of Reduce class and other utility functions
    """

    def __init__(
        self,
        x,
        f,
        params=None,
        C=None,
        g=None,
        h=None,
        u=None,
        params_values=None,
        x_init=None,
        timepoints_ode=None,
        timepoints_ssm=None,
        error_tol=None,
        nstates_tol=None,
    ):
        super().__init__(
            x,
            f,
            params,
            C,
            g,
            h,
            params_values,
            x_init,
            timepoints_ode,
            timepoints_ssm,
            error_tol,
            nstates_tol,
        )

    def write_results(self, filename):
        """
        Write the model reduction results in a file given by filename.
        The symbolic data is written in LaTeX.
        """
        from sympy.printing import latex

        f1 = open(filename, "w")
        f1.write("Model reduction results.\n")
        for key, value in self.results_dict.items():
            f1.write("A possible reduced model: \n \n")
            f1.write("\n$x_{hat} = ")
            f1.write(str(key.x))
            f1.write("$\n\n\n\n")
            for k in range(len(key.f)):
                f1.write("\n$f_{hat}(" + str(k + 1) + ") = ")
                f1.write(latex(key.f[k]))
                f1.write("$\n\n")
            f1.write("\n\n\n")
            f1.write("\nError metric:")
            f1.write(str(value[0]))
            f1.write("\n\n\n")
            f1.write("\nRobustness metric:")
            f1.write(str(value[1]))
            f1.write("\n\n\n")
            f1.write("Other properties")
            f1.write("\n\n\n")
            f1.write("\n C = ")
            f1.write(str(key.C))
            f1.write("\n$ g = ")
            f1.write(str(key.g))
            f1.write("$\n h = ")
            f1.write(str(key.h))
            f1.write("\n$h = ")
            f1.write(str(key.h))
            f1.write("$\n Solutions : \n")
            f1.write(str(key.x_sol))
            f1.write("\n\n\n\n")
            f1.write("\n Sensitivity Solutions : \n")
            f1.write(str(key.S))
            f1.write("\n\n\n\n")
        f1.close()

    def get_valid_reduced_models(self, nstates_tol=None, error_tol=None):
        """
        Returns the reduced models obtained and
        stored in results_dict that satisfy the given
        tolerances for number of states and the
        error tolerance. Among the return reduced
        model objects you may access robustness metric
        for each by looking at reduced_sys.Se.
        Choose the reduced model with lowest robustness metric.
        """
        if nstates_tol is None:
            nstates_tol = self.nstates_tol
        if error_tol is None:
            error_tol = self.error_tol
        valid_reduced_models = []
        results_dict = self.results_dict
        for key, value in results_dict.items():
            error = value[0]
            if error <= error_tol and len(key.x) <= nstates_tol:
                valid_reduced_models.append(key)
        self.valid_reduced_models = valid_reduced_models
        return valid_reduced_models


def create_system(
    x,
    f,
    params=None,
    C=None,
    g=None,
    h=None,
    u=None,
    params_values=None,
    x_init=None,
):
    return System(
        x,
        f=f,
        params=params,
        C=C,
        g=g,
        h=h,
        u=u,
        params_values=params_values,
        x_init=x_init,
    )



"""All model import/export functions"""

import numpy as np  # type: ignore
from sympy import Symbol, Integer  # type: ignore
from libsbml import (
    readSBMLFromFile,
    LIBSBML_SEV_FATAL,
    ConversionProperties,
    LIBSBML_OPERATION_SUCCESS,
)

from .model_reduction import Reduce


def load_ODE_model(n_states, n_params=0):
    """Directly load ODE with sympy"""
    return ode_to_sympy(n_states, n_params)


def ode_to_sympy(odesize, n_params=0):
    """
    Returns Sympy object for the given ODE function
    """
    from sympy import symbols  # type: ignore

    f = []
    x = []
    P = []
    for i in range(odesize):
        str_var = "x" + str(i)
        str_f = "f" + str(i)
        vars()[str_f] = symbols("f%d" % i)
        vars()[str_var] = symbols("x%d" % i)
        f.append(vars()[str_f])
        x.append(vars()[str_var])
    for k in range(n_params):
        str_P = "P" + str(k)
        vars()[str_P] = symbols("P" + "%d" % k)
        P.append(vars()[str_P])
    return x, f, P


def sympy_to_sbml(model):
    sbml_doc = None
    return sbml_doc


# SBML to ODE #
# This file reads an SBML file using libSBML,
#
# - expands all function definitions
# - expands all initial assignments
# - converts local parameter to global ones
# - then it goes ahead and write the ODE system
#   for use with scipy.integrate
# - it emits a function called simulateModel
#   that takes three parameters: t0, tend and numpoints
#   with that the model can be simulated as needed
# - finally the emitted function is called and the result plotted
# - it is also written out into a file called generated.py
#


def load_sbml(filename, **kwargs):
    """A function that takes in an SBML file and returns x,f,P,params_values.
    x is a list of species written as Sympy objects
    f is a list of functions written as Sympy objects
    P is a list of parameters written as Sympy objects
    params_values is a list of parameter values, in the same order as P
    x_init is a list of initial conditions, in the same order as x

    Returns: A reducible Reduce(System) object
    """

    # Get the sbml file, check for errors, and perform conversions
    doc = readSBMLFromFile(filename)
    if doc.getNumErrors(LIBSBML_SEV_FATAL):
        print("Encountered serious errors while reading file")
        print(doc.getErrorLog().toString())
        return
    doc.getErrorLog().clearLog()
    # Convert local params to global params
    props = ConversionProperties()
    props.addOption("promoteLocalParameters", True)
    if doc.convert(props) != LIBSBML_OPERATION_SUCCESS:
        print("The document could not be converted")
        print(doc.getErrorLog().toString())
    # Expand initial assignments
    props = ConversionProperties()
    props.addOption("expandInitialAssignments", True)
    if doc.convert(props) != LIBSBML_OPERATION_SUCCESS:
        print("The document could not be converted")
        print(doc.getErrorLog().toString())
    # Expand functions definitions
    props = ConversionProperties()
    props.addOption("expandFunctionDefinitions", True)
    if doc.convert(props) != LIBSBML_OPERATION_SUCCESS:
        print("The document could not be converted")
        print(doc.getErrorLog().toString())
    # Get model and define important lists, dictionaries
    mod = doc.getModel()
    x = []
    x_init = []
    P = []
    params_values = []
    reactions = {}
    # Append species symbol to 'x' and append initial
    # amount/concentration to x_init
    # x[i] corresponds to x_init[i]
    for i in range(mod.getNumSpecies()):
        species = mod.getSpecies(i)
        x.append(Symbol(species.getId()))
        if species.isSetInitialConcentration():
            x_init.append(species.getInitialConcentration())
        elif species.isSetInitialAmount():
            x_init.append(species.getInitialAmount())
        else:
            x_init.append(0)
    # Append parameter symbol to 'P' and parameter values to 'params_values'
    for i in range(mod.getNumParameters()):
        params = mod.getParameter(i)
        params_values.append(params.getValue())
        P.append(Symbol(params.getId()))
    # Get kinetic formula for each reaction, store in dictionary 'reactions'
    for i in range(mod.getNumReactions()):
        reaction = mod.getReaction(i)
        kinetics = reaction.getKineticLaw()
        # Convert species and parameter names to Symbols before sympify
        formula = kinetics.getFormula()
        for species in mod.getListOfSpecies():
            formula = formula.replace(
                species.getId(), f"Symbol('{species.getId()}')"
            )
        for param in mod.getListOfParameters():
            formula = formula.replace(
                param.getId(), f"Symbol('{param.getId()}')"
            )
        reactions[reaction.getId()] = eval(formula)
    # Define f
    f = [Integer(0)] * len(x)
    # Loop to define functions in 'f'
    for i in range(mod.getNumReactions()):
        reaction = mod.getReaction(i)
        # subtract reactant kinetic formula
        for j in range(reaction.getNumReactants()):
            ref = reaction.getReactant(j)
            species = Symbol(ref.getSpecies())
            curr_index = x.index(species)
            if ref.getStoichiometry() == 1.0:
                f[curr_index] += -reactions[reaction.getId()]
            else:
                f[curr_index] += (
                    -reactions[reaction.getId()] * ref.getStoichiometry()
                )
        # add product kinetic formula
        for j in range(reaction.getNumProducts()):
            ref = reaction.getProduct(j)
            species = Symbol(ref.getSpecies())
            curr_index = x.index(species)
            if ref.getStoichiometry() == 1.0:
                f[curr_index] += +reactions[reaction.getId()]
            else:
                f[curr_index] += (
                    +reactions[reaction.getId()] * ref.getStoichiometry()
                )
    if "outputs" in kwargs:
        outputs = kwargs["outputs"]
        if not isinstance(outputs, list):
            outputs = [outputs]
        C = np.zeros((len(outputs), len(x)))
        output_count = 0
        for output in outputs:
            index_output = x.index(Symbol(output))
            C[output_count, index_output] = 1
            output_count += 1
    else:
        C = None
    sys = Reduce(
        x,
        f,
        params=P,
        params_values=params_values,
        x_init=x_init,
        C=C,
        **kwargs,
    )
    return sys



# Import required libraries and dependencies
from sympy import Symbol
import libsbml
import numpy as np
from .sbmlutil import create_sbml_model, add_species, add_reaction
from .sbmlutil import add_parameters
from warnings import warn


class System(object):
    """
    Class that stores the system model in this form:
    x_dot = f(x, theta), y = Cx.
    """

    def __init__(
        self,
        x,
        f,
        params=None,
        C=None,
        g=None,
        h=None,
        u=None,
        params_values=None,
        x_init=None,
        input_values=None,
        **kwargs,
    ):
        """
        The general system dynamics :
        x_dot = f(x, P) + g(x, P)u, y = h(x,P)
        Use the utility function ode_to_sympy to write these.

        x : (Symbolic) state variable vector

        f : The system model dynamics.
            Written symbolically with symbols x = [x_0, x_1, ...]
            for states and P = [P_0, P_1, ...] for parameters.

        params : (Symbolic) parameters used to
                define f, g, h. None if no symbolic parameters.

        g : The actuator / input dynamics.
            None by default if the system is autonomous.

        C : The output matrix for y = Cx,
            size of C must be #outputs times #states. If None,
            the argument h is expected. Cannot set C and h both.

        h : The output description y = h(x, P)
            where x are states and P are parameters.

        params_values : Values for model parameters

        u : List of inputs

        x_init : Model initial conditions
        """

        self.x = x
        self.n = len(x)
        self.f = f
        self.params = params
        self.C = C
        self.g = g
        self.h = h
        self.u = u
        if params_values is not None:
            self.params_values = params_values
        else:
            self.params_values = []
        if input_values is not None:
            self.input_values = input_values
        else:
            self.input_values = []
        if x_init is not None:
            self.x_init = x_init
        else:
            self.x_init = []
        if "parameter_dependent_ic" in kwargs:
            self.parameter_dependent_ic = kwargs.get("parameter_dependent_ic")
        else:
            self.parameter_dependent_ic = False
        if "ic_parameters" in kwargs and kwargs.get("ic_parameters"):
            self.ic_parameters = kwargs.get("ic_parameters")
            if self.parameter_dependent_ic is False:
                raise ValueError(
                    "Make sure to set parameter_dependent_ic \
                    argument to True to use parameters as initial conditions"
                )
            if self.x_init == []:
                self.x_init = [] * len(self.ic_parameters)
            elif isinstance(self.x_init, np.ndarray):
                self.x_init = list(self.x_init)
            for ic_p, ic_i in zip(self.ic_parameters, range(len(self.x_init))):
                self.set_ic_from_params(self.x_init, ic_p, ic_i)
        else:
            self.ic_parameters = None
        return

    def set_dynamics(
        self, f=None, g=None, h=None, C=None, u=None, params=None
    ):
        """
        Set either f, g, h, or C to the
        System object or parameter values using P.
        """
        if f:
            self.f = f
        if g:
            self.g = g
        if h:
            self.h = h
        if C:
            self.C = C
        if u:
            self.u = u
        if params is not None:
            self.params = params
        else:
            self.params = []
        return self

    # Change this variable name P to params_values or something
    def evaluate(self, f, x, P, u=None, **kwargs):
        """
        Evaluate the given symbolic
        function (f) that is part of the System
        at the values given by x for self.x
        and P for self.params
        """
        if "set_params_as" in kwargs:
            set_params_as = kwargs["set_params_as"]
        else:
            set_params_as = None
        fs = []
        for i in range(len(f)):
            fi = f[i]
            fi = fi.subs(list(zip(self.x, x)))
            if self.u is not None:
                fi = fi.subs(list(zip(self.u, u)))
            if not set_params_as:
                fi = fi.subs(list(zip(self.params, P)))
            fs.append(fi)
        return fs

    def set_parameters(self, params_values=None, x_init=None):
        """
        Set model parameters and initial conditions
        """
        f_new = []
        if params_values is not None:
            self.params_values = [pi for pi in params_values]
            if self.params:
                for fi in self.f:
                    f_new.append(
                        fi.subs(list(zip(self.params, self.params_values)))
                    )
        else:
            self.params_values = []
        if x_init is not None:
            self.x_init = [pi for pi in x_init]
        else:
            self.x_init = []
        self.f = f_new
        return f_new

    def set_ic_from_params(self, x_init, ic_param, ic_index):
        """
        Set System initial conditions using parameter values
        """
        if isinstance(ic_param, Symbol):
            param_index = self.params.index(ic_param)
            value_from_params = self.params_values[param_index]
            x_init[ic_index] = value_from_params
            return value_from_params
        elif isinstance(ic_param, (int, float)):
            x_init[ic_index] = ic_param
            return ic_param
        else:
            raise ValueError(
                "Sympy Symbol or float expected in ic_parameters\
                             argument."
            )

    def generate_sbml_model(self, show_warnings=True, **kwargs):
        """Creates an new SBML model and populates with the species and
        their ODE in the System object
        :param show_warnings: bool, to display warnings
        :param kwargs: extra keywords pass onto create_sbml_model()
        :return: tuple: (document,model) SBML objects
        """
        document, model = create_sbml_model(**kwargs)
        states = [str(i) for i in self.x]
        states_ic = [float(i) for i in self.x_init]
        all_rxn_ids = [f"r{i}" for i in range(len(self.f))]
        params = [str(i) for i in self.params]
        for species, ic, ode_i, r_id in zip(
            states, states_ic, self.f, all_rxn_ids
        ):
            # Create species and initial conditions
            model = add_species(model, species, ic)
            # Reactions, for all species, s_i, --> s_i with
            # rate equal to the ODE term
            model = add_reaction(
                model,
                species=species,
                kinetic_law=ode_i,
                reaction_id=r_id,
                all_species=states,
            )
        model = add_parameters(
            model, all_parameters=params, all_values=self.params_values
        )
        if document.getNumErrors() and show_warnings:
            warn(
                "SBML model generated has errors."
                "Use document.getErrorLog() to print all errors."
            )
        return document, model

    def write_sbml(self, filename: str, **kwargs):
        """
        Writes an SBML file for AutoReduce System model.
        :param filename: String of file name to write SBML model to
        returns:
        :return SBMLDocument: libSBML SBMLDocument object
        """
        document, model = create_sbml_model(**kwargs)
        if "show_warnings" in kwargs:
            show_warnings = kwargs.get("show_warnings")
        else:
            show_warnings = True
        document, _ = self.generate_sbml_model(
            show_warnings=show_warnings, **kwargs
        )
        sbml_string = libsbml.writeSBMLToString(document)
        with open(filename, "w") as f:
            f.write(sbml_string)
        return document

    def __eq__(self, other):
        """
        Compare two System objects for equality.
        Two systems are equal if they have the same:
        - states (x)
        - dynamics (f)
        - parameters (params)
        - parameter values (params_values)
        - initial conditions (x_init)
        - output matrix (C)
        - input dynamics (g)
        - output description (h)
        - inputs (u)
        - input values (input_values)
        """
        if not isinstance(other, System):
            return False

        # Compare basic attributes
        if (
            self.x != other.x
            or self.f != other.f
            or self.params != other.params
            or self.params_values != other.params_values
            or self.x_init != other.x_init
            or self.C != other.C
            or self.g != other.g
            or self.h != other.h
            or self.u != other.u
            or self.input_values != other.input_values
        ):
            return False

        return True
